"""Module to run the freva deployment."""

from __future__ import annotations

import json
import logging
import os
import platform
import random
import shutil
import string
import sys
import time
from base64 import b64encode
from copy import deepcopy
from getpass import getuser
from pathlib import Path
from socket import gethostbyname, gethostname
from typing import Any, Dict, Optional, Union, cast
from urllib.parse import urlparse
from urllib.request import urlretrieve

import appdirs
import namegenerator
import tomlkit
import yaml
from rich import print as pprint
from rich.prompt import Prompt
from typing_extensions import NotRequired, TypedDict

import freva_deployment.callback_plugins
from freva_deployment import AUX_URL, FREVA_PYTHON_VERSION, __version__

from .error import ConfigurationError, handled_exception
from .keys import RandomKeys
from .logger import logger
from .runner import RunnerDir
from .utils import (
    RichConsole,
    asset_dir,
    config_dir,
    get_cache_information,
    get_passwd,
    is_localhost,
    load_config,
)
from .versions import get_steps_from_versions, get_versions

LOCAL_DEBUG_MSG = (
    "A freva test installation was successfully set up on your "
    "computer!\n\nYou can visit the freva-web app at "
    "https://localhost:8443\n"
    "You can log on with user: [b]{user}[/b], password: [b]{password}[/b]"
    "\n\nTo comepletly wipe the installation use the "
    "following commands:\n\n"
    "\tsystemctl stop --user 'freva-* data-loader* {project_name}-*' && \\ \n"
    "\trm -fr ~/.config/systemd/user/freva-*.service && \\ \n"
    "\trm -fr ~/.config/systemd/user/data-loader* && \\ \n"
    "\trm -fr ~/.config/systemd/user/{project_name}* && \\ \n"
    "\tsystemctl daemon-reload --user && \\ \n"
    "\trm -fr ~/.cache/freva-local\n\n"
    "You can omit the [i]--debug[/i] flag for a real deploymet."
)


class ConfigType(TypedDict):
    """Define the type of the yaml config dict."""

    hosts: NotRequired[str]
    vars: NotRequired[dict[str, str | int | bool | float]]


def get_current_architecture() -> str:
    """
    Determines the current architecture and operating system
    and maps it to one of the supported architectures.

    Supported architectures:
    - linux-64
    - linux-aarch64
    - linux-ppc64le
    - linux-s390x
    - osx-64
    - osx-arm64

    Returns:
        str: The corresponding architecture for mamba installation.
    """
    system = platform.system().lower()
    machine = platform.machine().lower()

    if system == "linux":
        if machine == "x86_64":
            return "linux-64"
        if machine in ("aarch64", "arm64"):
            return "linux-aarch64"
        if machine == "ppc64le":
            return "linux-ppc64le"
        if machine == "s390x":
            return "linux-s390x"
    elif system == "darwin":
        if machine == "x86_64":
            return "osx-64"
        if machine == "arm64":
            return "osx-arm64"
    return "linux-64"


def mamba_to_conda_arch(arch: str) -> str:
    """Translate a mamba to a conda arch."""
    translate = {
        "linux-64": "Linux-x86_64",
        "linux-aarch64": "Linux-aarch64",
        "linux-ppc64le": "Linux-ppc64le",
        "linux-s390x": "Linux-s390x",
        "osx-64": "MacOSX-x86_64",
        "osx-arm64": "MacOSX-arm64",
    }
    if arch in translate.values():
        return arch
    return translate.get(arch, "linux-64")


class DeployFactory:
    """Apply freva deployment and its services.

    Parameters
    ----------
    project_name: str
        The name of the project to distinguish this instance from others.
    steps: list[str], default: ["core", "web", "freva_rest", "db"]
        The components that are going to be deployed.
    config_file: os.PathLike, default: None
        Path to any existing deployment configuration file.
    local_debug: bool, default: False
        Run deployment only on local machine, debug mode.
    gen_keys: bool, default: False
        Create new set of certificates, if they don't already exist.

    Examples
    --------

    >>> from freva_deployment import DeployFactory as DF
    >>> deploy = DF(steps=["freva_rest"])
    >>> deploy.play(ask_pass=True)
    """

    step_order: tuple[str, ...] = (
        "core",
        "db",
        "freva_rest",
        "web",
    )
    _steps_with_cert: tuple[str, ...] = ("core", "db", "web")

    @handled_exception
    def __init__(
        self,
        steps: list[str] | None = None,
        config_file: Path | str | None = None,
        local_debug: bool = False,
        gen_keys: bool = False,
        _cowsay: bool = False,
    ) -> None:
        self.passwords: dict[str, str] = {}
        self._no_cowsay = str(int(_cowsay is False))
        self.gen_keys = gen_keys or local_debug
        self.local_debug = local_debug
        self._config_keys: list[str] = []
        self._master_pass: str = ""
        self._td: RunnerDir = RunnerDir()
        self.eval_conf_file: Path = self._td.parent_dir / "evaluation_system.conf"
        self.web_conf_file: Path = self._td.parent_dir / "freva_web.toml"
        self._db_pass: str = ""
        self._steps = steps or ["db", "freva_rest", "web", "core"]
        if self._steps in (["auto"], "auto"):
            self._steps = []
        self._inv_tmpl = Path(config_file or config_dir / "inventory.toml")
        self._cfg_tmpl = self.aux_dir / "evaluation_system.conf.tmpl"
        self.cfg = self._read_cfg()
        self.project_name = self.cfg.pop("project_name", None)
        if not self.project_name:
            raise ConfigurationError("You must set a project name") from None
        self._random_key = RandomKeys(
            self.project_name,
            self.cfg.get("web", {}).get("web_host", "localhost"),
        )
        self.current_step = ""
        self._check_steps_()

    def _check_steps_(self) -> None:
        """Before we do anything check if something is wrong with the config."""
        # First the steps that are needed
        old_master = self._master_pass
        self._master_pass = old_master or "foo"
        not_in_use = set(["db", "freva_rest", "web", "core"]) - set(self.steps)
        for step in self.steps:
            getattr(self, f"_prep_{step.replace('-', '_')}")()
        for step in not_in_use:
            try:
                getattr(self, f"_prep_{step.replace('-', '_')}")()
            except ConfigurationError as error:
                logger.warning("Some of your config isn't valid:\n%s", error)
        self._master_pass = old_master

    @property
    def master_pass(self) -> str:
        """Define the master password."""
        if self._master_pass:
            return self._master_pass
        self._master_pass = get_passwd(os.environ.get("MASTER_PASSWD", ""))
        return self._master_pass

    @property
    def public_key_file(self) -> str:
        """Path to the public certificate file."""
        public_keyfile = self.cfg["certificates"].get("public_keyfile")
        chain_keyfile = self.cfg["certificates"].get("chain_keyfile")
        if public_keyfile:
            return str(Path(public_keyfile).expanduser().absolute())
        if chain_keyfile:
            return str(Path(chain_keyfile).expanduser().absolute())
        if self.gen_keys:
            return self._random_key.public_chain_file
        raise ConfigurationError(
            "You must give a valid path to a public key file."
        ) from None

    @property
    def private_key_file(self) -> str:
        """Path to the private key file."""
        keyfile = self.cfg["certificates"].get("private_keyfile")
        if keyfile:
            return str(Path(keyfile).expanduser().absolute())
        if self.gen_keys:
            return self._random_key.private_key_file
        raise ConfigurationError(
            "You must give a valid path to a private key file."
        ) from None

    def _prep_vault(self) -> None:
        """Prepare the vault."""
        self._config_keys.append("vault")
        self.cfg["vault"].setdefault("ansible_become_user", "root")
        self.cfg["vault"].pop("db_playbook", "")
        self.cfg["vault"]["db_port"] = self.cfg["vault"].get("port", 3306)
        self.cfg["vault"]["version"] = __version__
        self.cfg["vault"]["db_user"] = self.cfg["vault"].get("user", "")
        self.cfg["vault"]["root_passwd"] = self.master_pass
        self.cfg["vault"]["passwd"] = self.db_pass
        self.cfg["vault"]["keyfile"] = self.public_key_file
        self.cfg["vault"]["email"] = self.cfg["web"].get("contacts", "")
        self.cfg["vault"]["db_host"] = self.cfg["db"]["db_host"]

    def _prep_db(self) -> None:
        """prepare the mariadb service."""
        self._config_keys.append("db")
        self.cfg["db"].setdefault("ansible_become_user", "root")
        host = self.cfg["db"]["db_host"]
        self.cfg["db"]["root_passwd"] = self.master_pass
        self.cfg["db"]["passwd"] = self.db_pass
        self.cfg["db"]["keyfile"] = self.public_key_file
        data_path = Path(
            cast(
                str,
                self.cfg["db"].get("data_path", "/opt/freva"),
            )
        )
        self.cfg["db"]["data_path"] = str(data_path)
        for key in ("name", "user", "db"):
            self.cfg["db"][key] = self.cfg["db"].get(key) or "freva"
        db_host = self.cfg["db"].get("host", "").strip()
        if not db_host:
            self.cfg["db"]["host"] = host
        self.cfg["db"].setdefault("port", "3306")
        self.cfg["db"]["email"] = self.cfg["web"].get("contacts", "")
        self._prep_vault()
        self._prep_web(False)

    def _prep_data_loader(self) -> None:
        """Prepare the data-loader."""
        data_path = Path(
            cast(
                str,
                self.cfg["freva_rest"].get("data_path", "/opt/freva"),
            )
        )
        redis_host = (
            self.cfg["freva_rest"].get("redis_host", "")
            or self.cfg["freva_rest"]["freva_rest_host"]
        )
        data_portal_host = self.cfg["freva_rest"].get(
            "data_loader_portal_hosts", ""
        )
        data_portal_hosts = [
            d.strip() for d in data_portal_host.split(",") if d.strip()
        ]
        if data_portal_hosts:
            scheduler_host = data_portal_hosts[0]
        else:
            scheduler_host = ""
        for prefix in ("redis://", "http://", "https://", "tcp://"):
            redis_host = redis_host.removeprefix(prefix)
            scheduler_host = scheduler_host.removeprefix(prefix)

        redis_host, _, redis_port = redis_host.partition(":")
        redis_port = redis_port or "6379"
        self.cfg["freva_rest"][
            "redis_cache_url"
        ] = f"redis://{redis_host}:{redis_port}"
        self.cfg["freva_rest"]["redis_host_name"] = redis_host
        proxy_url = (
            f'http://{self.cfg["freva_rest"]["freva_rest_host"]}:'
            f'{self.cfg["freva_rest"]["freva_rest_port"]}'
        )
        proxy_url = (
            self.cfg["web"].get("project_website", "").strip() or proxy_url
        )
        scheme, _, netloc = proxy_url.rpartition("://")
        scheme = scheme or "http"
        self.cfg["freva_rest"]["proxy_url"] = f"{scheme}://{netloc}"
        scheduler_host, _, scheduler_port = scheduler_host.partition(":")
        redis_information = get_cache_information(
            redis_host,
            redis_port,
            scheduler_host,
            scheduler_port or "40000",
        )
        redis_information["passwd"] = self._create_random_passwd(30, 10)
        redis_information_enc = b64encode(
            json.dumps(redis_information).encode("utf-8")
        ).decode("utf-8")
        self._config_keys.append("redis")
        self._config_keys.append("data_portal_hosts")
        self._config_keys.append("data_portal_scheduler")
        if not scheduler_host:
            self.cfg["data_portal_scheduler"] = self.cfg["redis"] = self.cfg[
                "data_portal_hosts"
            ] = {}
            return
        # Update the config.
        self.cfg["data_portal_scheduler"] = {
            "data_portal_scheduler_host": scheduler_host or "",
            "information": redis_information_enc,
            "admin_user": self.cfg["freva_rest"].get("admin_user", ""),
            "is_worker": False,
            "ansible_become_user": self.cfg["freva_rest"].get(
                "ansible_become_user", "root"
            ),
            "ansible_user": self.cfg["freva_rest"].get("ansible_user", getuser()),
        }
        if data_portal_hosts[1:]:
            self.cfg["data_portal_hosts"] = {
                "data_portal_hosts": ",".join(data_portal_hosts[1:]),
                "information": redis_information_enc,
                "admin_user": self.cfg["freva_rest"].get("admin_user", ""),
                "ansible_become_user": self.cfg["freva_rest"].get(
                    "ansible_become_user", "root"
                ),
                "ansible_user": self.cfg["freva_rest"].get(
                    "ansible_user", getuser()
                ),
            }
        self.cfg["redis"] = {
            "redis_host": redis_host,
            "information": redis_information_enc,
            "project_name": self.project_name,
            "port": redis_port,
            "admin_user": self.cfg["freva_rest"].get("admin_user", ""),
            "data_path": str(data_path),
            "ansible_become_user": self.cfg["freva_rest"].get(
                "ansible_become_user", "root"
            ),
            "ansible_user": self.cfg["freva_rest"].get("ansible_user", getuser()),
            "information": redis_information_enc,
        }

    def _prep_freva_rest(self, prep_web=True) -> None:
        """prepare the freva_rest service."""
        self._config_keys.append("freva_rest")
        self._prep_data_loader()
        self.cfg["freva_rest"].setdefault("ansible_become_user", "root")
        self.cfg["freva_rest"]["root_passwd"] = self.master_pass
        self.cfg["freva_rest"]["db_passwd"] = self._create_random_passwd(30, 10)
        freva_rest_host = self.cfg["freva_rest"]["freva_rest_host"]
        vault_host = self.cfg["db"].get("vault_host") or self.cfg["db"]["db_host"]
        self.cfg["freva_rest"]["vault_host"] = vault_host
        self.cfg["freva_rest"]["db_host"] = freva_rest_host
        self.cfg["freva_rest"]["db_user"] = namegenerator.gen()
        self.cfg["freva_rest"].pop("core", None)
        services = ["databrowser", "stacapi"]
        if self.cfg["freva_rest"].get("data_loader_portal_hosts", "").strip():
            services.append("zarr-stream")
            self.cfg["freva_rest"]["data_loader"] = True
        else:
            self.cfg["freva_rest"]["data_loader"] = False
        data_path = Path(
            cast(
                str,
                self.cfg["freva_rest"].get("data_path", "/opt/freva"),
            )
        )
        self.cfg["freva_rest"]["data_path"] = str(data_path)
        for key, default in dict(solr_mem="4g", freva_rest_port=7777).items():
            self.cfg["freva_rest"][key] = (
                self.cfg["freva_rest"].get(key) or default
            )
        self.cfg["freva_rest"]["email"] = self.cfg["web"].get("contacts", "")
        self.cfg["freva_rest"].setdefault("oidc_url", "")
        self.cfg["freva_rest"].setdefault("oidc_client", "freva")
        self.cfg["freva_rest"].setdefault("oidc_client_secret", "")
        token_claims = self.cfg["freva_rest"].get("oidc_token_claims") or {}
        if isinstance(token_claims, str):
            self.cfg["freva_rest"]["oidc_token_claims"] = token_claims
        else:
            token_claims_str = []
            for key, values in token_claims.items():
                if isinstance(values, str):
                    values = [values]
                for v in values:
                    token_claims_str.append(f"{key}:{v}")
            self.cfg["freva_rest"]["oidc_token_claims"] = ",".join(
                token_claims_str
            )
        self.cfg["freva_rest"]["services"] = ",".join(services)
        if prep_web:
            self._prep_web(False)
        for key in ("mongodb_server", "search_server"):
            self._config_keys.append(key)
            self.cfg[key] = deepcopy(self.cfg["freva_rest"])

    def _prep_core(self) -> None:
        """prepare the core deployment."""
        self._config_keys.append("core")
        self.cfg["core"].setdefault("ansible_become_user", "")
        # Legacy args as we are going to use micromamba
        self.cfg["core"]["arch"] = mamba_to_conda_arch(
            self.cfg["core"]
            .get("arch", "linux-64")
            .lower()
            .replace("x86_", "")
            .replace("mac", "")
        )
        self.cfg["core"]["python_version"] = FREVA_PYTHON_VERSION
        self.cfg["core"]["admins"] = self.cfg["core"].get("admins") or getuser()
        if not self.cfg["core"]["admins"]:
            self.cfg["core"]["admins"] = getuser()
        install_dir = Path(self.cfg["core"]["install_dir"])
        root_dir = Path(
            self.cfg["core"].get("root_dir", "").strip() or install_dir
        )
        self.cfg["core"]["install_dir"] = str(install_dir)
        self.cfg["core"]["root_dir"] = str(root_dir)
        preview_path = self.cfg["core"].get("preview_path", "")
        base_dir_location = self.cfg["core"].get("base_dir_location", "") or str(
            root_dir / "work"
        )
        self.cfg["core"]["base_dir_location"] = base_dir_location
        scheduler_output_dir = self.cfg["core"].get("scheduler_output_dir", "")
        scheduler_system = self.cfg["core"].get("scheduler_system", "local")
        if not preview_path:
            self.cfg["core"]["preview_path"] = str(
                Path(base_dir_location) / "share" / "preview"
            )
        if not scheduler_output_dir:
            scheduler_output_dir = str(Path(base_dir_location) / "share")
        elif Path(scheduler_output_dir).parts[-1] != scheduler_system:
            scheduler_output_dir = Path(scheduler_output_dir) / scheduler_system
        self.cfg["core"]["scheduler_output_dir"] = str(scheduler_output_dir)
        self.cfg["core"]["keyfile"] = self.public_key_file
        self.cfg["core"]["git_url"] = "https://github.com/FREVA-CLINT/freva.git"

    def _prep_web(self, ask_pass: bool = True) -> None:
        """prepare the web deployment."""
        self._config_keys.append("web")

        for key in "oidc_url", "oidc_client", "oidc_client_secret":
            self.cfg["web"][key] = self.cfg["freva_rest"].get(key, "")
        self.cfg["web"].setdefault("ansible_become_user", "root")
        self._prep_core()
        freva_rest_host = (
            f'{self.cfg["freva_rest"]["freva_rest_host"]}:'
            f'{self.cfg["freva_rest"]["freva_rest_port"]}'
        )
        self.cfg["web"]["freva_rest_host"] = freva_rest_host
        self.cfg["web"].setdefault("deploy_web_server", True)
        data_path = Path(
            cast(
                str,
                self.cfg["web"].get("data_path", "/opt/freva"),
            )
        )
        self.cfg["web"]["data_path"] = str(data_path)
        admin = self.cfg["core"]["admins"]
        try:
            for k, v in self.cfg["web"].items():
                self.cfg["web"].setdefault(k, "")
        except KeyError:
            raise ConfigurationError(
                "No web config section given, please configure the web.config"
            ) from None
        if not isinstance(admin, str):
            self.cfg["web"]["admin"] = admin[0]
        else:
            self.cfg["web"]["admin"] = admin
        allowed_hosts = self.cfg["web"].get("allowed_hosts") or ["localhost"]
        if isinstance(allowed_hosts, str):
            allowed_hosts = [
                s.strip() for s in allowed_hosts.split(",") if s.strip()
            ]
        allowed_hosts.append(self.cfg["web"]["web_host"])
        allowed_hosts.append(f"{self.project_name}-httpd")
        self.cfg["web"]["allowed_hosts"] = [
            s.strip() for s in set(allowed_hosts) if s.strip()
        ]
        self.cfg["web"].setdefault("main_color", "Tomato")
        _webserver_items = {
            "institution_logo": self.cfg["web"].get("institution_logo", ""),
            "main_color": self.cfg["web"].get("main_color", "Tomato"),
            "border_color": self.cfg["web"].get("border_color", "#6c2e1f"),
            "hover_color": self.cfg["web"].get("hover_color", "#d0513a"),
            "homepage_text": self.cfg["web"].get("homepage_text", ""),
            "imprint": self.cfg["web"].get("imprint", []),
            "homepage_heading": self.cfg["web"].get("homepage_heading", ""),
            "about_us_text": self.cfg["web"].get("about_us_text", ""),
            "contacts": self.cfg["web"].get("contacts", []),
            "insitution_name": self.cfg["web"].get("insitution_name", ""),
            "menu_entries": self.cfg["web"].get("menu_entries", []),
        }
        try:
            with Path(_webserver_items["homepage_text"]).open("r") as f_obj:
                _webserver_items["homepage_text"] = f_obj.read()
        except (FileNotFoundError, IOError, KeyError):
            pass
        try:
            with Path(_webserver_items["about_us_text"]).open() as f_obj:
                _webserver_items["about_us_text"] = f_obj.read()
        except (FileNotFoundError, IOError, KeyError):
            pass
        web_config = tomlkit.dumps(_webserver_items)
        self.web_conf_file.write_text(web_config)
        self.cfg["web"]["config_content"] = b64encode(
            web_config.encode()
        ).decode()
        server_name = self.cfg["web"].pop("server_name", [])
        if isinstance(server_name, str):
            server_name = server_name.split(",")
        server_name = ",".join([a for a in server_name if a.strip()])
        if not server_name:
            server_name = self.cfg["web"]["web_host"]
        self.cfg["web"]["server_name"] = server_name
        web_host = self.cfg["web"]["web_host"]
        if web_host == "127.0.0.1":
            web_host = "localhost"
        self.cfg["web"]["host"] = web_host

        self.cfg["web"]["csrf_trusted_origins"] = []
        for url in [server_name] + self.cfg["web"]["project_website"].split(","):
            trusted_origin = urlparse(url.strip())

            if trusted_origin.scheme:
                netloc = trusted_origin.netloc.removeprefix("www.")
            else:
                netloc = trusted_origin.path.removeprefix("www.")
            for prefix in ("www.", ""):
                uri = f"https://{prefix}{netloc}"
                self.cfg["web"]["csrf_trusted_origins"].append(uri)
        host_ip = gethostbyname(web_host)
        self.cfg["web"]["csrf_trusted_origins"].append(f"http://{host_ip}")
        self.cfg["web"]["csrf_trusted_origins"].append(f"https://{host_ip}")

        self.cfg["web"]["csrf_trusted_origins"] = list(
            set(self.cfg["web"]["csrf_trusted_origins"])
        )
        self.cfg["web"]["freva_bin"] = os.path.join(
            self.cfg["core"]["install_dir"], "bin"
        )
        for key in ("core", "web"):
            self.cfg[key]["config_toml_file"] = str(self.web_conf_file)
        self._prep_vault()
        self.cfg["vault"]["ansible_python_interpreter"] = self.cfg["db"].get(
            "ansible_python_interpreter",
        )
        self.cfg["web"]["root_passwd"] = self.master_pass
        self.cfg["web"]["private_keyfile"] = self.private_key_file
        self.cfg["web"]["public_keyfile"] = self.public_key_file
        self.cfg["web"]["key_content"] = b64encode(
            Path(self.private_key_file).read_bytes()
        ).decode("utf-8")
        self.cfg["web"]["cert_content"] = b64encode(
            Path(self.public_key_file).read_bytes()
        ).decode("utf-8")
        self.cfg["web"]["eval_conf_file"] = str(
            Path(self.cfg["core"]["root_dir"])
            / "freva"
            / "evaluation_system.conf"
        )
        chatbot_host = self.cfg["web"].get("chatbot_host", "") or "localhost"
        self.cfg["web"]["vault_host"] = self.cfg["db"].get(
            "vault_host"
        ) or self.cfg["db"].get("db_host", "localhost")
        self.cfg["web"]["chatbot_host"] = chatbot_host
        if self._random_key.check_cert_key_pair(
            self.public_key_file, self.private_key_file
        ):
            self.cfg["web"]["chatbot_rest_api_url"] = f"https://{web_host}"
        else:
            self.cfg["web"]["chatbot_rest_api_url"] = f"http://{freva_rest_host}"
        self.cfg["web"]["redis_username"] = namegenerator.gen()
        self.cfg["web"]["redis_password"] = self._create_random_passwd()

    def _prep_local_debug(self, cfg: dict[str, Any]) -> dict[str, Any]:
        """Prepare the system for a potential local debug."""
        self._master_pass = "secret"
        default_ports = {"db": 3306, "freva_rest": 7777}
        cfg.setdefault("project_name", "freva-test-deployment")
        cfg["deployment_method"] = "conda"
        deploy_dir = Path(appdirs.user_data_dir("freva-local"))
        deploy_dir.mkdir(exist_ok=True, parents=True)
        host = gethostbyname(gethostname())
        cfg["deployment_dir"] = str(deploy_dir)
        steps = [k for (k, v) in cfg.items() if isinstance(v, dict)]
        for step in steps:
            for var in cfg[step]:
                if var.endswith("_host") or var.endswith("_hosts"):
                    cfg[step][var] = host
                if var.endswith("admin_user"):
                    cfg[step][var] = getuser()
            cfg[step]["wipe"] = False
            if step in default_ports:
                cfg[step]["port"] = default_ports[step]
            cfg[step]["ansible_user"] = getuser()
            cfg[step]["ansible_python_interpreter"] = sys.executable
            cfg[step]["ansible_become_user"] = ""
            cfg[step]["data_path"] = str(deploy_dir)
        cfg["db"]["db_host"] = cfg["db"]["vault_host"] = host
        cfg["core"]["ansible_become_user"] = ""
        cfg["core"]["scheduler_system"] = "local"
        cfg["core"]["admins"] = getuser()
        cfg["web"]["scheduler_host"] = ["localhost"]
        cfg["web"]["allowed_hosts"] = [host, "localhost", "127.0.0.1"]
        cfg["web"]["project_website"] = "https://localhost"
        cfg["core"]["arch"] = get_current_architecture()
        cfg["freva_rest"]["oidc_url"] = (
            "http://localhost:8080/realms/freva/"
            ".well-known/openid-configuration"
        )
        cfg["freva_rest"]["oidc_client"] = "freva"
        cfg["freva_rest"]["oidc_client_secret"] = ""
        cfg["freva_rest"]["oidc_token_claims"] = ""
        for key in (
            "root_dir",
            "base_dir_location",
            "preview_path",
            "scheduler_output_dir",
            "admin_group",
            "root_dir",
        ):
            cfg["core"][key] = ""
        cfg["core"]["install_dir"] = str(
            deploy_dir / cfg["project_name"] / "freva-core"
        )
        return cfg

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._td.cleanup()

    def _read_cfg(self) -> dict[str, Any]:
        mapper = {
            "vault": "db",
            "mongodb_server": "freva_rest",
            "search_server": "freva_rest",
        }
        try:
            config = dict(load_config(self._inv_tmpl, convert=True).items())
            for dest, source in mapper.items():
                host = (
                    config[source].get(f"{dest}_host")
                    or config[source][f"{source}_host"]
                )
                config[dest] = deepcopy(config[source])
                config[dest][f"{dest}_host"] = config[source][f"{dest}_host"] = (
                    host
                )
            if self.local_debug:
                config = self._prep_local_debug(config)
            return config
        except FileNotFoundError:
            raise ConfigurationError(f"No such file {self._inv_tmpl}") from None
        except KeyError as error:
            raise ConfigurationError(str(error)) from error

    def _check_config(self) -> None:
        sections = []
        config_keys = [s for s in set(self._config_keys)]
        for section in self.cfg.keys():
            for step in config_keys:
                if section.startswith(step) and section not in sections:
                    sections.append(section)
        for section in sections:
            for key, value in self.cfg[section].items():
                if (
                    not value
                    and not self._empty_ok
                    and not isinstance(value, bool)
                ):
                    raise ConfigurationError(
                        f"{key} in {section} is empty in {self._inv_tmpl}"
                    ) from None

    @property
    def ask_become_password(self) -> bool:
        """Check if we have to ask for the sudo passwd at all."""
        cfg = deepcopy(self.cfg)
        for step in ("freva_rest", "web", "db"):
            if isinstance(cfg[step], dict):
                ansible_become_user = cfg[step].get("ansible_become_user", "root")
                if ansible_become_user:
                    return True
        return False

    @property
    def _empty_ok(self) -> list[str]:
        """Define all keys that can be empty."""
        return [
            "admins",
            "conda_exec_path",
            "ansible_become_user",
        ]

    def _get_files_copy(self, key) -> Path | None:
        return dict(
            core=self.eval_conf_file.absolute(),
            web=self.eval_conf_file.absolute(),
        ).get(key, None)

    @staticmethod
    def _create_random_passwd(num_chars: int = 30, num_digits: int = 8) -> str:
        num_chars -= num_digits
        characters = [
            "".join(
                [random.choice(string.ascii_letters) for i in range(num_chars)]
            ),
            "".join([random.choice(string.digits) for i in range(num_digits)]),
        ]
        str_characters = "".join(characters)
        _db_pass = "".join(random.sample(str_characters, len(str_characters)))
        while _db_pass.startswith("@"):
            # Vault treats values starting with "@" as file names.
            _db_pass = "".join(random.sample(str_characters, len(str_characters)))
        return _db_pass

    @property
    def db_pass(self) -> str:
        """Create a password for the database."""
        if self._db_pass:
            return self._db_pass
        self._db_pass = self._create_random_passwd(30, 8)
        return self._db_pass

    @property
    def _needs_core(self) -> list[str]:
        """Define the steps that need the core config."""
        return ["web", "core"]

    def _set_additional_config_values(
        self,
        step: str,
        config: dict[str, ConfigType],
    ) -> None:
        """Set additional values to the configuration."""
        if step in self._needs_core:
            for key in (
                "root_dir",
                "base_dir_location",
                "preview_path",
                "scheduler_output_dir",
            ):
                value = self.cfg["core"].get(key, "")
                config[step]["vars"][f"core_{key}"] = value
        config[step]["vars"][f"{step}_name"] = f"{self.project_name}-{step}"
        config[step]["vars"]["asset_dir"] = str(asset_dir)
        config[step]["vars"]["ansible_user"] = (
            self.cfg[step].get("ansible_user") or getuser()
        )
        python_exe = self.cfg[step].get("ansible_python_interpreter", "")
        if python_exe.strip():
            config[step]["vars"][
                f"{step}_ansible_python_interpreter"
            ] = python_exe
            config[step]["vars"]["ansible_python_interpreter"] = python_exe

        dump_file = self._get_files_copy(step)
        if dump_file:
            config[step]["vars"][f"{step}_dump"] = str(dump_file)

    def parse_config(self, steps: list[str], **extra: str) -> Optional[str]:
        """Create config files for ansible and evaluation_system.conf."""
        versions = get_versions()
        additional_steps = set(steps) - set(self.steps)
        if additional_steps:
            pprint(
                "The following, [b]not selected[/b] steps will be auto "
                f"updated.\n[green]{', '.join(additional_steps)}[/]"
            )
            time.sleep(3)
        new_steps = set(steps + self.steps)
        if not new_steps:
            return None
        self._steps = list(new_steps)
        logger.info("Parsing configurations")
        self._check_config()
        _ = [getattr(self, f"_prep_{step}")() for step in self.steps]
        kube_conf = self.cfg.get("kubernetes", {})
        config: dict[str, ConfigType] = {
            "kubernetes": {
                "hosts": kube_conf.get("deploy_host", "localhost"),
                "vars": {
                    "expose_method": kube_conf.get("expose_method", "lb"),
                    "config_only": kube_conf.get("config_only", False),
                    "project_name": self.project_name,
                    "ansible_become_user": self.cfg["freva_rest"].get(
                        "become_user"
                    )
                    or "root",
                    "ansible_user": self.cfg["freva_rest"].get("ansible_user")
                    or getuser(),
                    "base_path": self.cfg["freva_rest"].get("data_path")
                    or "/opt/freva",
                    "deployment_method": "k8s",
                    "service": "",
                },
            }
        }

        playbooks = "\n".join(
            [
                a.read_text()
                for a in (asset_dir / "playbooks").rglob("*.*")
                if a.is_file()
            ]
        )
        main_playbook = yaml.safe_load(
            (asset_dir / "playbooks" / "main-deployment.yml").read_text()
        )
        tags, hosts = [], []
        for entry in main_playbook:
            for tag in entry.get("tags", []):
                if tag in self.steps + ["vault"]:
                    tags.append(entry["hosts"])
        no_prepend = ("root_passwd", "deployment_method", "expose_method")
        for step in set(self._config_keys):
            config[step] = {}
            if self.cfg.get(step, {}).get(f"{step}_host"):
                config[step]["hosts"] = self.cfg[step][f"{step}_host"]
            elif self.cfg.get(step, {}).get(step):
                config[step]["hosts"] = self.cfg[step][step]
            else:
                continue
            if step != "core" and step in config:
                hosts.append(config[step]["hosts"])
            config[step]["vars"] = {}
            for key, value in self.cfg[step].items():
                if key in no_prepend + (f"{step}_host",) or key.startswith(step):
                    new_key = key
                else:
                    new_key = f"{step.replace('-', '_')}_{key}"
                if new_key in playbooks:
                    config[step]["vars"][new_key] = value
            config[step]["vars"]["project_name"] = self.project_name
            config[step]["vars"][f"{step}_admin_user"] = self.cfg[step].get(
                "admin_user", ""
            )
            config[step]["vars"]["deployment_method"] = self.cfg.get(
                "deployment_method", "docker"
            )
            if step in versions:
                config[step]["vars"][f"{step.replace('-', '_')}_version"] = (
                    versions[step]
                )
            config[step]["vars"]["debug"] = self.local_debug
            # Add additional keys
            self._set_additional_config_values(step, config)
            for k, v in extra.items():
                config[step]["vars"][k] = v
        max_width = int(max(shutil.get_terminal_size().columns * 0.75, 25))
        if "search_server" in config:
            config["search_server"]["vars"]["solr_version"] = versions["solr"]
        if "db" in config:
            config["db"]["vars"]["vault_version"] = versions["vault"]
            for key, value in config["vault"]["vars"].items():
                if key.startswith("vault_"):
                    config["db"]["vars"].setdefault(key, value)
        if "web" in config:
            config["web"]["vars"]["proxy_version"] = versions.get("nginx", "")
        info: Dict[str, Any] = {}
        for step in config:
            for key, value in config[step].get("vars", {}).items():
                if key in extra:
                    continue
                if step in tags or ("db" in self.steps and step == "vault"):
                    info.setdefault(step, {})
                    info[step].setdefault("vars", {})
                    add_value: Union[str, int, float] = value
                    if isinstance(value, str) and "pass" in key:
                        add_value = "***"
                    elif isinstance(value, str) and len(value) > max_width:
                        add_value = value[:max_width] + "..."
                    info[step]["vars"][key] = add_value
        info_str = yaml.dump(json.loads(json.dumps(info)))
        for passwd in (self.master_pass,):
            if passwd:
                info_str = info_str.replace(passwd, "***")
        if logger.level > logging.DEBUG:
            RichConsole.print(info_str, style="bold", markup=True)
        core_host: Optional[str] = config.get("core", {}).get("hosts")
        if core_host in hosts:
            config["core"]["hosts"] = gethostbyname(core_host) or ""
        return yaml.dump(json.loads(json.dumps(config)))

    @property
    def python_prefix(self) -> Path:
        """Get the path of the new conda evnironment."""
        return Path(sys.exec_prefix) / "bin" / "python3"

    @property
    def aux_dir(self) -> Path:
        """Directory with auxiliary files."""
        return asset_dir / "config"

    @property
    def steps(self) -> list[str]:
        """Set all the deployment steps."""
        return [s for s in self.step_order if s in self._steps]

    def _set_deployment_methods(self) -> str:
        """Check the deployment methods."""
        valid_deployment_methods = ("podman", "docker", "conda", "k8s")
        deployment_method = self.cfg.get("deployment_method", "")
        if deployment_method not in valid_deployment_methods:
            raise ConfigurationError(
                f"Deployment method: {deployment_method} is invalid, should be"
                f"one of {', '.join(valid_deployment_methods)}"
            )
        return deployment_method

    def create_eval_config(self) -> Optional[Path]:
        """Create and dump the evaluation_system.config."""
        logger.info("Creating evaluation_system.conf")
        keys = (
            ("core", "admins"),
            ("web", "project_website"),
            ("core", "root_dir"),
            ("core", "base_dir_location"),
            ("core", "preview_path"),
            ("core", "scheduler_output_dir"),
            ("core", "scheduler_system"),
        )
        cfg_file = asset_dir / "config" / "evaluation_system.conf.tmpl"
        if not cfg_file.is_file():
            cfg_file.parent.mkdir(exist_ok=True, parents=True)
            urlretrieve(AUX_URL, filename=str(cfg_file))

        with cfg_file.open() as f_obj:
            lines = f_obj.readlines()
            for num, line in enumerate(lines):
                if line.startswith("project_name"):
                    lines[num] = f"project_name={self.project_name}\n"
                for key, value in keys:
                    if line.startswith(value):
                        cfg = self.cfg[key].get(value, "")
                        if cfg:
                            lines[num] = f"{value}={cfg}\n"
                for step in ("freva_rest", "db"):
                    cfg = self.cfg[step].get("port", "")
                    if line.startswith(f"{step}.port"):
                        lines[num] = f"{step}.port={cfg}\n"
                    if line.startswith(f"{step}.host"):
                        lines[num] = (
                            f"{step}.host={self.cfg[step][f'{step}_host']}\n"
                        )
                if line.startswith("solr.host"):
                    lines[num] = f"solr.host={self.cfg[step][f'{step}_host']}\n"
                if line.startswith("db.host"):
                    lines[num] = f"db.host={self.cfg['db']['db_host']}\n"
        dump_file = self._get_files_copy("core")
        if dump_file:
            with dump_file.open("w") as f_obj:
                f_obj.write("".join(lines))
        return dump_file

    def get_ansible_password(self, ask_pass: bool = False) -> dict[str, str]:
        """The passwords for the ansible environments."""
        ssh_pass_msg = "Give the [b]ssh[/b] password for remote login"
        sudo_pass_msg = "Give the password elevating user privilege ([b]sudo[/b])"
        if ask_pass:
            sudo_pass_msg += ", defaults to ssh password"
        passwords = {}
        ssh_key = "ansible_ssh_pass"
        sudo_key = "ansible_become_pass"
        passwords[sudo_key] = os.environ.get("ANSIBLE_BECOME_PASSWORD", "") or ""
        passwords[ssh_key] = os.environ.get("ANSIBLE_SSH_PASSWORD", "") or ""
        if ask_pass and not passwords[ssh_key]:
            passwords[ssh_key] = Prompt.ask(
                f"[green]{ssh_pass_msg}[/green]", password=True
            )
        if not passwords[sudo_key] and self.ask_become_password:
            passwords[sudo_key] = Prompt.ask(
                f"[green]{sudo_pass_msg}[/green]", password=True
            )
            if not passwords[sudo_key]:
                passwords[sudo_key] = passwords.get(ssh_key, "")
        return passwords

    @handled_exception
    def play(
        self,
        ask_pass: bool = True,
        verbosity: int = 0,
        ssh_port: int = 22,
        skip_version_check: bool = False,
        tags: Optional[list[str]] = None,
        local: bool = False,
        extra: Optional[Dict[str, str]] = None,
    ) -> None:
        """Play the ansible playbook.

        Parameters
        ----------
        ask_pass: bool, default: True
            Instruct Ansible to ask for the ssh password instead of using a
            ssh key
        verbosity: int, default: 0
            Verbosity level, default 0
        ssh_port: int, default: 22
            Set the ssh port, in 99.9% of cases this should be left at port 22
        skip_version_check: bool, default: False
            Skip the version check, use with caution
        tags: list[str], default: None
            Instead of running the steps, fine grain the deployment using this
            specific tasks.
        local: bool, default: False
            Use local connections only.
        extra:
            Add/Override inventory settings.

        """
        try:
            self._play(
                ask_pass=ask_pass,
                verbosity=verbosity,
                ssh_port=ssh_port,
                skip_version_check=skip_version_check,
                tags=tags,
                extra=extra,
            )
        except KeyboardInterrupt as error:
            if str(error):
                pprint(f" [red][ERROR]: {error}[/]", file=sys.stderr)
            raise KeyboardInterrupt() from None

    def get_steps_from_versions(
        self,
        envvars: dict[str, str],
        extravars: dict[str, str],
        passwords: dict[str, str],
        verbosity: int,
    ) -> list[str]:
        """Check the versions of the different freva parts."""
        config: dict[str, ConfigType] = {}
        versions = get_versions()
        steps = list(set(self.step_order) - set(self.steps))
        if not steps or self.local_debug:
            # The user has selected all steps anyway, nothing to do here:
            return []
        playbook_tmpl = yaml.safe_load(
            (asset_dir / "playbooks" / "versions.yaml").read_text()
        )
        playbook = []
        cfg = deepcopy(self.cfg)
        if "web" in steps and not cfg.get("web", {}).get("web_host", "").strip():
            steps = [s for s in steps if s != "web"]
        steps.append("vault")
        version_path = self._td.parent_dir / "versions.txt"
        hosts = []
        for tasks in playbook_tmpl:
            step = tasks["hosts"]
            if step in steps:
                playbook.append(tasks)
                host_var = cfg[step][f"{step}_host"]
                if step != "core":
                    hosts.append(host_var)
                config[step] = {}
                config[step]["hosts"] = host_var
                ansible_user = cfg[step].get("ansible_user") or os.getenv(
                    "ANSIBLE_USER", getuser()
                )
                if "ansible_become_user" not in cfg[step]:
                    become_user = "root"
                else:
                    become_user = cfg[step]["ansible_become_user"]

                config[step]["vars"] = {
                    f"{step}_ansible_become_user": become_user,
                    "asset_dir": str(asset_dir),
                    "deployment_method": self.cfg.get(
                        "deployment_method", "docker"
                    ),
                    f"{step}_ansible_user": ansible_user,
                    "project_name": self.project_name,
                    f"{step}_data_path": cfg[step].get("data_path", ""),
                    "version_file_path": str(version_path),
                    f"{step.replace('-', '_')}_version": versions[step],
                }
                python_exe = self.cfg[step].get("ansible_python_interpreter", "")
                config[step]["vars"][f"{step}_ansible_python_interpreter"] = (
                    python_exe or "/usr/bin/python"
                )
        config.setdefault("core", {})
        config["core"].setdefault("vars", {})
        config["core"]["vars"]["core_install_dir"] = cfg["core"]["install_dir"]
        core_host: Optional[str] = config.get("core", {}).get("hosts")
        if core_host and core_host in hosts:
            config["core"]["hosts"] = gethostbyname(core_host) or ""
        result = self._td.run_ansible_playbook(
            playbook=playbook,
            inventory=config,
            envvars=envvars,
            passwords=passwords,
            extravars=extravars,
            verbosity=verbosity,
            hide_output=True,
            text="Getting versions of micro-services ...",
        )
        versions = {}
        for line in result.splitlines():
            jline = json.loads(line)
            if "msg" in jline["result"] and jline["task"].lower().startswith(
                "display"
            ):
                service = jline["task"].split()[1].lower()
                version = jline["result"]["msg"].strip()
                versions[service.strip()] = version.strip()
        logger.debug("Detected versions: %s", versions)
        additional_steps = get_steps_from_versions(versions)
        return additional_steps

    def _play(
        self,
        ask_pass: bool = True,
        verbosity: int = 0,
        ssh_port: int = 22,
        skip_version_check: bool = False,
        tags: Optional[list[str]] = None,
        local: bool = False,
        extra: Optional[Dict[str, str]] = None,
    ) -> None:
        extra = extra or {}
        plugin_path = Path(freva_deployment.callback_plugins.__file__).parent
        envvars: dict[str, str] = {
            "ANSIBLE_CONFIG": str(self._td.ansible_config_file),
            "ANSIBLE_NOCOWS": self._no_cowsay.lower(),
            "ANSIBLE_COW_PATH": os.getenv(
                "ANSIBLE_COW_PATH", shutil.which("cowsay") or ""
            ),
            "ANSIBLE_ACTION_WARNINGS": str(int(verbosity > 0)),
            "ANSIBLE_DEVEL_WARNING": str(int(verbosity > 0)),
            "ANSIBLE_CALLBACK_RESULT_FORMAT": "yaml",
            "ANSIBLE_CALLBACK_FORMAT_PRETTY": "true",
        }
        config = {
            "cowsay_enabled_stencils": "default,sheep,moose",
            "stdout_callback": "deployment_plugin",
            "callback_result_format": "yaml",
            "result_format": "yaml",
            "callback_format_pretty": "True",
            "callback_plugins": str(plugin_path),
            "host_key_checking": "False",
            "retry_files_enabled": "False",
            "nocows": str(bool(int(self._no_cowsay))).lower(),
            "action_warnings": str(verbosity > 0).lower(),
            "devel_warning": str(verbosity > 0).lower(),
            "cowpath": os.getenv(
                "ANSIBLE_COW_PATH", shutil.which("cowsay") or ""
            ),
            "cow_selection": "random",
            "interpreter_python": "auto_silent",
            "timeout": "15",
        }
        self._td.create_config(**config)
        logger.debug("CONFIG\n%s", self._td.ansible_config_file.read_text())
        extravars: dict[str, str] = {
            "ansible_port": str(ssh_port),
            "ansible_config": str(self._td.ansible_config_file),
            "ansible_ssh_args": "-o ForwardX11=no -o StrictHostKeyChecking=no",
        }

        self.passwords = self.get_ansible_password(ask_pass)
        self._set_deployment_methods()
        local_connection = local or self.local_debug
        steps = [s for s in self.steps]
        if skip_version_check is False:
            steps = list(
                set(
                    steps
                    + self.get_steps_from_versions(
                        envvars.copy(),
                        extravars.copy(),
                        self.passwords.copy(),
                        verbosity,
                    )
                )
            )
        tags = [t for t in tags or steps]
        ask_master_pass = False
        for t in tags:
            if t in ["vault", "db", "web"]:
                ask_master_pass = True
        if ask_master_pass is False:
            self._master_pass = "foo"
        if local_connection:
            extravars["ansible_connection"] = "local"
        inventory = self.parse_config(steps, **extra)
        if inventory is None:
            logger.info("Services up to date, nothing to do!")
            return None
        if self.local_debug:
            logger.info("Overriding configuration for local deployment!")
        logger.debug(inventory)
        self.create_eval_config()
        RichConsole.rule(
            f"[r]Playing tasks: [i]{', '.join(tags or steps)}[/] with ansible[/]"
        )
        time.sleep(3)
        config.pop("stdout_callback")
        config.pop("callback_plugins")
        self._td.create_config(**config)
        self._td.run_ansible_playbook(
            working_dir=asset_dir,
            playbook=asset_dir / "playbooks" / "main-deployment.yml",
            inventory=inventory,
            envvars=envvars,
            tags=list(set(tags)),
            passwords=self.passwords,
            extravars=extravars,
            verbosity=verbosity,
        )
        if self.local_debug:
            RichConsole.rule("[b red]:bulb:   NOTE:[/]")
            RichConsole.print(
                LOCAL_DEBUG_MSG.format(
                    project_name=self.project_name,
                    user=getuser(),
                    password=self.master_pass,
                )
            )
            RichConsole.rule("")



"""Command line interface for the deployment."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import List, Optional

from rich_argparse import ArgumentDefaultsRichHelpFormatter

from freva_deployment import __version__

from ..deploy import DeployFactory
from ..error import DeploymentError
from ..logger import set_log_level
from ..utils import config_dir
from ..versions import VersionAction, display_versions


class BatchParser:
    """Command line interface for batchmode deployment."""

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        epilog: str = "",
    ) -> None:
        self.parser = parser or argparse.ArgumentParser(
            prog="deploy-freva-cmd",
            description="Deploy freva and its services on different machines.",
            formatter_class=ArgumentDefaultsRichHelpFormatter,
            epilog=epilog,
        )
        self.parser.add_argument(
            "--config",
            "-c",
            type=Path,
            help="Path to ansible inventory file.",
            default=config_dir / "config" / "inventory.toml",
        )
        self.parser.add_argument(
            "--steps",
            "-s",
            type=str,
            nargs="+",
            default=["db", "freva-rest", "web", "core"],
            choices=["web", "core", "db", "freva-rest", "auto"],
            help=(
                "The services/code stack to be deployed. Use [b]auto[/b]"
                " to only deploy outdated services"
            ),
        )
        self.parser.add_argument(
            "--ask-pass",
            help="Connect to server via ssh passwd instead of public key.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--ssh-port",
            help="Set the ssh port, in 99.9%% of the cases this should be 22",
            type=int,
            default=22,
        )
        self.parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            help="Verbosity level",
            default=0,
        )
        self.parser.add_argument(
            "-l",
            "--local",
            help="Deploy services on the local machine.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--debug",
            help="Deploy services on the local machine.",
            action="store_true",
            default=False,
        )

        self.parser.add_argument(
            "-g",
            "--gen-keys",
            help="Generate public and private web certs, use with caution.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--skip-version-check",
            help="Skip the version check. Use with caution.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action=VersionAction,
            version="[b][red]%(prog)s[/red] {version}[/b]{services}".format(
                version=__version__, services=display_versions()
            ),
        )
        self.parser.add_argument(
            "-t",
            "--tags",
            type=str,
            nargs="+",
            default=None,
            choices=[
                "core",
                "db",
                "cache",
                "data-loader",
                "freva-rest",
                "mongodb",
                "search-server",
                "pre-web",
                "web",
            ],
            help=(
                "Fine grain deployment. Instead of using steps you can "
                "set those ansible tasks ([i]tags[/i]) to be deployed."
            ),
        )
        self.parser.add_argument(
            "-e",
            "--extra",
            type=str,
            nargs=2,
            action="append",
            help="Add/Override inventory settings.",
        )
        self.parser.add_argument(
            "--cowsay",
            action="store_true",
            help="Let the cow speak!",
            default=False,
        )
        self.parser.set_defaults(cli=self.run_cli)

    def parse_args(self, argv: Optional[List[str]]) -> argparse.Namespace:
        return self.parser.parse_args()

    @staticmethod
    def run_cli(args: argparse.Namespace) -> None:
        """Run the command line interface."""
        set_log_level(args.verbose)
        steps = [s.replace("-", "_") for s in args.steps]
        extra = dict(args.extra) if args.extra else None
        if args.tags:
            steps = []
        with DeployFactory(
            steps=steps,
            config_file=args.config,
            local_debug=args.debug,
            gen_keys=args.gen_keys,
            _cowsay=args.cowsay,
        ) as DF:
            try:
                DF.play(
                    args.ask_pass,
                    args.verbose,
                    ssh_port=args.ssh_port,
                    skip_version_check=args.skip_version_check,
                    tags=args.tags or None,
                    local=args.local,
                    extra=extra,
                )
            except KeyboardInterrupt:
                raise SystemExit(130)
            except DeploymentError:
                raise SystemExit(1)


def cli(argv: list[str] | None = None) -> None:
    """Run the command line interface."""
    epilog = (
        "[red][b]Note:[/b] The command `deploy-freva-cmd` is a legacy command,"
        " please consider using `deploy-freva cmd` instead.[/red]"
    )
    bp = BatchParser(epilog=epilog)
    args = bp.parse_args(argv)
    bp.run_cli(args)


if __name__ == "__main__":
    cli(sys.argv[1:])



"""Download and extract cowsay binary from an archive."""

import argparse
import os
import tarfile
import tempfile
import zipfile
from pathlib import Path
from urllib.request import urlretrieve


def reporthook(count, block_size, total_size):
    if count == 0:
        return
    frac = count * block_size / total_size
    percent = int(100 * frac)
    bar = "#" * int(frac * 40)
    msg = "Downloading: [{0:<{1}}] | {2}% Completed".format(bar, 40, round(percent, 2))
    print(msg, end="\r", flush=True)
    if frac >= 1:
        print()


class CowsayExtractor:
    def __init__(self, url: str):
        self.url = url
        self.script_dir = Path(__file__).parent
        self.temp_file = None

    def __enter__(self):
        """Create a temporary file and prepare for extraction."""
        self.temp_file = tempfile.NamedTemporaryFile(
            delete=False, suffix=Path(self.url).suffix
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close and delete the temporary file."""
        if self.temp_file:
            self.temp_file.close()

    def download_file(self) -> Path:
        """Download the archive from the specified URL.

        Returns
        -------
        Path
            The path to the downloaded archive.
        """
        urlretrieve(self.url, filename=self.temp_file.name, reporthook=reporthook)
        return Path(self.temp_file.name)

    def extract_cowsay(self, archive_path: Path) -> Path:
        """Extract the cowsay binary from a ZIP or TAR.GZ archive.

        Parameters
        ----------
        archive_path : Path
            The path to the archive file from which to extract the binary.

        Returns
        -------
        Path: Path to extracted executable.
        """
        if archive_path.suffix == ".zip":
            with zipfile.ZipFile(archive_path, "r") as archive:
                for member in archive.namelist():
                    if member in ["cowsay", "cowsay.exe"]:
                        archive.extract(member, path=self.script_dir)
                        extracted_path = self.script_dir / member
                        break
        elif archive_path.suffix == ".gz":
            with tarfile.open(archive_path, "r:gz") as archive:
                for member in archive.getmembers():
                    if member.name in ["cowsay", "cowsay.exe"]:
                        archive.extract(member, path=self.script_dir)
                        extracted_path = self.script_dir / member.name
                        break
        else:
            raise ValueError("Unsupported archive format. Must be .zip or .tar.gz.")
        if extracted_path.suffix != ".exe" and os.name != "nt":
            try:
                os.chmod(extracted_path, 0o755)
            except OSError:
                print(f"Failed to set executable bit on {extracted_path}")

    def run(self) -> None:
        """Main function to download and extract the cowsay binary."""
        archive_path = self.download_file()
        self.extract_cowsay(archive_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Download and extract cowsay binary from an archive."
    )
    parser.add_argument(
        "url",
        type=str,
        help="URL of the ZIP or TAR.GZ file containing cowsay.",
    )
    args = parser.parse_args()

    with CowsayExtractor(args.url) as extractor:
        extractor.run()



#!/usr/bin/env python3
import argparse
import asyncio
import hashlib
import os
import platform
import shutil
import socket
import subprocess
import sys
import tarfile
import tempfile
import time
import urllib.request as req
from functools import wraps
from getpass import getuser
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import (
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    ParamSpec,
    Type,
    TypeVar,
)

F = TypeVar("F", bound=Callable)
P = ParamSpec("P")

SERVICE = """[Unit]
Description=Automated freva deployments
After=network.target
Wants=network.target
After=network.target
[Service]
Type=simple
NoNewPrivileges=true
SendSIGKILL=no
KillSignal=SIGTERM
PermissionsStartOnly=true
ExecStart={prefix}/conda/bin/python {prefix}/automation-setup.py
StateDirectory=prefect
Environment="HOME=/var/lib/prefect"
Environment="PREFECT_HOME=/var/lib/prefect"
Environment="PREFECT_API_DATABASE_CONNECTION_URL=sqlite+aiosqlite:////var/lib/prefect/prefect.db?timeout=30"
Environment="PREFECT_API_URL=http://127.0.0.1:8888/api"
StandardOutput=journal
StandardError=journal
StateDirectory=prefect
Environment="PATH={prefix}/conda/bin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin"
EnvironmentFile=/etc/sysconfig/freva-automation.conf
WorkingDirectory={prefix}
StandardOutput=journal
StandardError=journal
Restart=on-failure
RestartSec=5
# User=<set user for security>
# Group=<set group for security>

[Install]
WantedBy=multi-user.target
"""


CONFIG = """
freva = "{path}"
"""


def load_env_configs() -> List[Path]:
    env_var = "FREVA_DEPLOYMENT_CONFIG"
    return [
        Path(f.strip()) for f in os.getenv(env_var, "").split(",") if f.strip()
    ]


def mock_decorator(func: Callable[P, F]) -> Callable[P, F]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> F:
        return func(*args, **kwargs)

    return wrapper


try:
    import appdirs
    import toml
    from prefect import flow, get_run_logger, task
    from prefect.client.orchestration import get_client
    from prefect.client.schemas.filters import (
        DeploymentFilter,
        DeploymentFilterName,
    )
    from prefect.exceptions import FailedRun
    from prefect.filesystems import LocalFileSystem

    bootstrap = False
except ImportError as error:
    print(f"[WARNING] Module not found: {error}")
    print("[WARNING] Boot strapping")
    bootstrap = True
    task = flow = mock_decorator


class BootstrapConda:
    """Install conda into the prefix location."""

    def call_from_bootstrap(self):
        """Call the script in a bootstrap env."""
        with tempfile.NamedTemporaryFile(suffix=".py") as temp_f:
            Path(temp_f.name).write_text(sys.stdin.read())
            cmd = [self.executable, temp_f.name]
            env = os.environ.copy()
            env.pop("_CALLED_FROM_BOOTSTRAP", None)
            subprocess.check_call(cmd, env=env)

    def __init__(
        self, prefix: Path, extra_pkgs: Optional[List[str]] = None
    ) -> None:
        extra_pkgs = extra_pkgs or []
        prefix = prefix.expanduser().absolute()
        micromamba = prefix / "bin" / "micromamba"
        self.prefix = prefix
        pkgs = list(
            set(
                ["appdirs", "prefect", "caddy", "toml", "python=3.12", "git"]
                + extra_pkgs
            )
        )
        if not micromamba.is_file():
            self._setup_micromamba()
        conda_env = prefix / "conda"
        self.executable = str(prefix / "conda" / "bin" / "python")
        if not conda_env.is_dir():
            subprocess.check_call(
                [
                    f"{micromamba}",
                    "create",
                    "-y",
                    "-p",
                    f"{conda_env}",
                    "-c",
                    "conda-forge",
                    "--override-channels",
                ]
                + pkgs
            )
            subprocess.check_call([self.executable, "-m", "ensurepip"])

    def _setup_micromamba(self):
        system = platform.system().lower()
        plt = platform.machine()
        target = self.prefix / "bin" / "micromamba"
        if system != "linux":
            raise SystemExit("Only Linux based deployment is supported.")
        if plt.startswith("x86"):
            plt = "64"
        with NamedTemporaryFile(suffix=".tar") as temp_f:
            self._url_retrieve(
                f"https://micro.mamba.pm/api/micromamba/linux-{plt}/latest",
                temp_f.name,
            )
            with tarfile.open(temp_f.name, mode="r:bz2") as tar:
                member = tar.getmember("bin/micromamba")
                print("🔧 Extracting: bin/micromamba")
                tar.extract(member, path=str(self.prefix))
        target.chmod(0o755)

    def _url_retrieve(self, conda_url: str, target: str) -> None:
        print("Retrieving conda install script: {}".format(conda_url))
        req.urlretrieve(
            str(conda_url),
            filename=target,
        )


__version__ = "0.1.0"
SCRIPT_URL = "https://raw.githubusercontent.com/freva-org/freva-admin/refs/heads/main/automation/automation-setup.py"
SCRIPT_CHECKSUM = (
    "c2bcfa7702a46d753748fea4c8a21daf2c79b32ef924cef280c1c53d6f9fdf89"
)

Steps = Literal["core", "freva-rest", "db", "web"]


def project_type() -> Type:

    try:
        import toml
    except ImportError:
        return Literal["freva"]
    cfg_file = Path(__file__).parent / "freva-automation.toml"
    projects = list(toml.load(cfg_file).keys())
    if not projects:
        return Literal["freva"]
    return Literal[*projects]


def get_env_vars() -> Dict[str, str]:
    envs = dict(os.environ.copy())
    if envs.get("FREVA_AUTOMATION_ENV_FILE"):
        content = Path(envs["FREVA_AUTOMATION_ENV_FILE"]).read_text()
        for line in content.splitlines():
            key, _, value = line.partition("=")
            if key.strip() and value.strip():
                envs[key.strip()] = value.strip()
    envs["INTERACTIVE_DEPLOY"] = "0"
    envs["PATH"] = os.path.dirname(sys.executable) + os.pathsep + envs["PATH"]
    envs["MASTER_PASSWD"] = envs["FREVA_AUTOMATION_PASSWORD"]
    return envs


@task
def update_deployment_software():
    return
    subprocess.check_output(
        [sys.executable, "-m", "pip", "install", "-U", "freva-deployment"]
    )


@task
def run_scripts(script_path: Path, prefix: str = "*"):

    logger = get_run_logger()
    for path in script_path.glob(f"{prefix}.sh"):
        try:
            PrefectServer.execute(["sh", str(path.absolute())])
        except subprocess.CalledProcessError as error:
            logger.error("Could not execute script: %s", path)


@task
def run_deployment(
    config_file: Path,
    steps: List[Steps],
    gen_certs: bool = False,
    use_ssh_pass: bool = False,
    verbosity: int = 0,
    skip_version_check: bool = False,
    extra: Optional[List[str]] = None,
):
    extra = extra or []
    steps = steps or ["auto"]
    cmd = ["deploy-freva", "cmd", "-c", f"{config_file}"]
    if verbosity:
        cmd.append("-" + "v" * verbosity)
    if use_ssh_pass:
        cmd.append("--ask-pass")
    if gen_certs:
        cmd.append("-g")
    if skip_version_check:
        cmd.append("--skip-version-check")
    cmd += ["-s"] + steps
    for ex in extra:
        key, _, value = ex.partition("=")
        if key and value:
            cmd += ["-e", key.strip(), value.strip()]

    PrefectServer.execute(cmd)


@flow
def freva_deployment_flow(
    project: project_type(),
    steps: Optional[List[Steps]] = None,
    gen_certs: bool = False,
    verbosity: int = 0,
    use_ssh_pass: bool = False,
    skip_version_check: bool = False,
    extra: Optional[List[str]] = None,
):
    """
    Apply the freva deployment for a Project.

    Args:
        project: The freva project that should be deployed.
        steps: The steps that should be run.
        gen_certs: Auto generate self signed keys.
        verbosity: Set the verbosity level.
        skip_version_check: Skip checking the micro server versions.
        use_ssh_pass: Instead of trying to connect via ssh keys, use a ssh password
        extra: Any config items that should be overridden.
    """
    update_deployment_software()
    if os.getenv("SCRIPT_PATH"):
        run_scripts(Path(os.getenv("SCRIPT_PATH")))
    steps = steps or ["auto"]
    config = toml.loads(
        (Path(__file__).parent / "freva-automation.toml").read_text()
    )
    run_deployment(
        config[project],
        steps,
        gen_certs,
        use_ssh_pass,
        verbosity,
        skip_version_check=skip_version_check,
        extra=extra,
    )


def get_user_config_dir(appname: str) -> Path:
    if sys.platform.startswith("win"):
        base = os.getenv("APPDATA", Path.home() / "AppData" / "Roaming")
    elif sys.platform == "darwin":
        base = Path.home() / "Library" / "Application Support"
    else:
        base = Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config"))
    cfg_dir = Path(base) / appname
    cfg_dir.mkdir(exist_ok=True, parents=True)
    return cfg_dir


class PrefectServer:
    """Start the prefect server."""

    @staticmethod
    def execute(cmd):
        """Execute a command and log the stdout to the prefect logger."""

        logger = get_run_logger()
        logger.info("Running : %s", " ".join(cmd))
        env = get_env_vars()
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
            universal_newlines=True,
            env=env,
        )
        assert proc.stdout is not None
        while True:
            line = proc.stdout.readline()
            if not line and proc.poll() is not None:
                break
            s_line = (line or "").rstrip().strip()
            if s_line:
                logger.info(s_line)
        if proc.returncode != 0:
            stderr_output = proc.stderr.read()
            logger.error(
                "Process %s exited with code : %i", " ".join(cmd), proc.returncode
            )
            logger.error("Captured stderr:\n%s", stderr_output.strip())
            raise FailedRun(f"{' '.join(cmd)} exited: {proc.returncode}")

    def __init__(
        self,
        log_dir: Path,
        api_port: int,
        web_port: Optional[int] = None,
        endpoint: Optional[str] = None,
    ) -> None:

        self.caddy_bin = shutil.which(
            "caddy", path=str(Path(sys.executable).parent)
        )
        self.username = os.getenv("FREVA_AUTOMATION_USERNAME")
        self.password = os.getenv("FREVA_AUTOMATION_PASSWORD")
        if not self.username or not self.password:
            raise SystemExit(
                "FREVA_AUTOMATION_USERNAME and FREVA_AUTOMATION_PASSWORD must be set"
            )

        if not self.caddy_bin:
            raise SystemExit("Caddy bin not found.")
        if web_port is None:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("", 0))
                self.port = s.getsockname()[1]
        else:
            self.port = web_port
        self.api_port = api_port
        log_dir.mkdir(exist_ok=True, parents=True)

        self.caddy_dir = Path(tempfile.mkdtemp(prefix="caddy-config-"))
        self.log_dir = log_dir
        self.server_log = (log_dir / "prefect-server.log").open("w")
        self.agent_log = (log_dir / "prefect-agent.log").open("w")
        self.caddy_log = (log_dir / "caddy.log").open("w")
        self.env = os.environ.copy()
        self.caddy_host = f"{socket.gethostname()}:{self.api_port}"
        self.prefect_api_url = f"http://127.0.0.1:{self.port}/api"
        self.env["PREFECT_API_URL"] = self.prefect_api_url
        if endpoint:
            self.env["PREFECT_UI_API_URL"] = (
                f"https://{self.caddy_host}/{endpoint}/api"
            )
        else:
            self.env["PREFECT_UI_API_URL"] = f"https://{self.caddy_host}/api"

    def linger(self) -> None:
        """Wait until the server gets terminated."""
        try:
            while True:
                server_status = self.server.poll()
                agent_status = self.agent.poll()
                if server_status is not None or agent_status is not None:
                    break
                time.sleep(2)
        except (KeyboardInterrupt, SystemExit):
            pass
        self.server.terminate()
        self.agent.terminate()
        self.server_log.close()
        self.agent_log.close()

    def prep_server(
        self, script_path: Path, user_var: Optional[str] = None
    ) -> None:
        """Prepare the server startup."""
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "freva-deployment"]
        )
        from freva_deployment.utils import config_file

        if user_var:
            config_dir = Path(appdirs.user_config_dir("freva")) / "deployment"
            config_dir.mkdir(exist_ok=True, parents=True)
            config = {"general": {}, "variables": {"USER": user_var}}
            config_file.write_text(toml.dumps(config))

        if not script_path.is_dir():
            print("[WARNING] Script path is not a directory.")
            return
        for path in script_path.glob("*.sh"):
            try:
                subprocess.run(
                    ["sh", str(path.absolute())],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    check=True,
                    env=get_env_vars(),
                )
            except subprocess.CalledProcessError as error:
                print(f"[WARNING] Failed to execute {path}: {error.stderr}")

    def start_caddy(
        self,
        cert_file: Optional[Path] = None,
        key_file: Optional[Path] = None,
    ) -> None:
        """Create and launch Caddy reverse proxy with basic auth."""
        if cert_file is None or key_file is None:
            from freva_deployment.keys import RandomKeys

            keys = RandomKeys("freva-automation", socket.gethostname())
            key_file = Path("freva-automation.key").absolute()
            cert_file = Path("freva-automation.crt").absolute()
            key_file.write_text(Path(keys.private_key_file).read_text())
            cert_file.write_text(Path(keys.public_chain_file).read_text())

        # Generate bcrypt hash
        hashed_password = subprocess.check_output(
            [self.caddy_bin, "hash-password", "--plaintext", self.password],
            text=True,
        ).strip()
        # Create temp dir for Caddy config
        caddyfile = self.caddy_dir / "Caddyfile"
        caddyfile.write_text(
            f"""{{
    auto_https disable_redirects

}}

{self.caddy_host} {{
    reverse_proxy 127.0.0.1:{self.port}
    tls {cert_file} {key_file}
    basicauth / {{
        {self.username} {hashed_password}
    }}
}}

:8080 {{
    redir https://{self.caddy_host}{{uri}} permanent
}}"""
        )
        subprocess.check_call(
            [self.caddy_bin, "fmt", "--overwrite", "--config", str(caddyfile)],
            env=self.env,
        )
        self.caddy = subprocess.Popen(
            [
                self.caddy_bin,
                "run",
                "--config",
                str(caddyfile),
                "--adapter",
                "caddyfile",
            ],
            stdout=self.caddy_log,
            stderr=self.caddy_log,
            env=self.env,
        )
        print("[INFO] Reverse proxy started visit: " f"https://{self.caddy_host}")

    def stop_server(self) -> None:
        """Stop all services."""
        for attr in ("server", "agent", "caddy"):
            proc = getattr(self, attr, None)
            if proc:
                proc.terminate()
            log = getattr(self, f"{attr}_log", None)
            if log:
                log.close()

    def start_prefect(self) -> None:
        """Start the prefect server."""
        subprocess.check_call(
            [sys.executable, "-m", "prefect", "server", "stop"], env=self.env
        )
        print("[INFO] Starting Prefect server...")
        self.server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "prefect",
                "server",
                "start",
                "--port",
                f"{self.port}",
                "--host",
                "127.0.0.1",
            ],
            stdout=self.server_log,
            stderr=self.server_log,
            env=self.env,
        )
        time.sleep(10)
        print("[INFO] Starting Prefect agent...")
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "prefect",
                "work-pool",
                "create",
                "--overwrite",
                "--type",
                "process",
                "default-agent-pool",
            ],
            env=self.env,
        )
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "prefect",
                "block",
                "register",
                "-m",
                "prefect.filesystems",
            ],
            env=self.env,
        )
        self.agent = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "prefect",
                "worker",
                "start",
                "--pool",
                "default-agent-pool",
            ],
            env=self.env,
            stdout=self.agent_log,
            stderr=self.agent_log,
        )
        time.sleep(5)
        print(f"[INFO] Prefect web app started at: http://localhost:{self.port}")


def write_config(config: dict, output_path: Path):
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_path, "w") as stream:
        toml.dump(config, stream)


async def delete_existing_deployment(name: str):
    """Delete existing deployments."""
    async with get_client() as client:
        deployments = await client.read_deployments(
            deployment_filter=DeploymentFilter(
                name=DeploymentFilterName(any_=[name])
            )
        )
        for dep in deployments:
            await client.delete_deployment(dep.id)
            print(f"[INFO] Deleted existing deployment: {dep.id}")


def register_prefect_deployment(
    name: str = "Freva Deployments",
) -> None:
    basepath = Path(__file__).parent.absolute()

    LocalFileSystem(
        basepath=str(basepath),
    ).save(name="freva-local", overwrite=True)
    LocalFileSystem.load("freva-local")

    this_file = Path(__file__).name
    asyncio.run(delete_existing_deployment(name))
    freva_deployment_flow.from_source(
        source=Path(__file__).parent,
        entrypoint=f"{this_file}:freva_deployment_flow",
    ).deploy(
        name=name,
        tags=["Deployment"],
        work_pool_name="default-agent-pool",
    )


# --- Self-update & checksum logic ---
def self_update(script_url: str, destination: Path):

    print(f"[INFO] Downloading latest script from: {script_url}")
    with req.urlopen(script_url) as response:
        content = response.read()
        destination.write_bytes(content)
    print(f"[INFO] Script updated at: {destination}")
    os.chmod(destination, 0o755)


def calculate_checksum(path: Path) -> str:
    with open(path, "rb") as f:
        return hashlib.sha256(f.read()).hexdigest()


# --- CLI
def main():
    parser = argparse.ArgumentParser(
        description="Prepare Freva config and register Prefect.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--prefix",
        "-p",
        help="The path where the automation should be deployed to.",
        type=Path,
        default=Path(
            os.getenv("FREVA_AUTOMATION_PREFIX_DIR") or Path(__file__).parent
        ),
    )
    parser.add_argument(
        "--config",
        "-c",
        nargs="*",
        help="Path(s) to the deployment files.",
        default=load_env_configs(),
    )
    parser.add_argument(
        "--version", action="store_true", help="Show script version and exit"
    )
    parser.add_argument(
        "--self-update",
        action="store_true",
        help="Fetch and replace script with latest version",
    )
    parser.add_argument(
        "--verify-checksum",
        action="store_true",
        help="Verify SHA256 checksum of script",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=os.getenv("FREVA_AUTOMATION_SERVER_PORT", "4200"),
        help="Set the port to run the server.",
    )
    parser.add_argument(
        "--extra-pkg",
        type=str,
        default=os.getenv("FREVA_AUTOMATION_EXTRA_PGKS") or None,
        help="Extra conda-forge packages that need to be installed. "
        "',' separated.",
    )
    parser.add_argument(
        "--script-directory",
        type=Path,
        default=os.getenv("FREVA_AUTOMATION_SCRIPT_DIR", "."),
        help="The directory that contains any pre-processing scripts "
        "that should be executed before launching the automation "
        "server.",
    )
    parser.add_argument(
        "--user",
        type=str,
        default=os.getenv("ANSIBLE_USER") or getuser(),
        help="If the deployment config file defines the `USER` variable. "
        "you can set the value of this variable here.",
    )
    parser.add_argument(
        "--endpoint",
        type=str,
        default=os.getenv("FREVA_AUTOMATION_SERVER_ENDPOINT") or None,
        help="Set the reverse proxy endpoint that serves the web app.",
    )
    parser.add_argument(
        "--cert-file",
        type=Path,
        default=os.getenv("FREVA_AUTOMATION_CERT_FILE") or None,
        help="Path to the web server certificate file.",
    )
    parser.add_argument(
        "--key-file",
        type=Path,
        default=os.getenv("FREVA_AUTOMATION_KEY_FILE") or None,
        help="Path to the web server private key file.",
    )
    parser.add_argument(
        "--web-port",
        type=int,
        default=os.getenv("FREVA_AUTOMATION_WEB_PORT") or None,
        help="Port the web app should be running on. This is useful if you "
        "don't want to spawn a revers porxy but use an existing one.",
    )
    parser.add_argument(
        "--no-reverse-proxy",
        action="store_true",
        help="Do not start a caddy reverse proxy.",
    )
    args = parser.parse_args()
    port = int(args.port)
    script_path = Path(__file__).resolve()
    extra_pkgs = [
        pkg.strip() for pkg in (args.extra_pkg or "").split(",") if pkg.strip()
    ]
    prefix = args.prefix.absolute().expanduser()
    if bootstrap or not (prefix / "conda" / "bin" / "prefect").is_file():
        service = prefix / "freva-automation.service"
        config = prefix / "freva-automation.toml"
        args.prefix.mkdir(exist_ok=True, parents=True)
        script_path = prefix / "automation-setup.py"
        if not script_path.is_file() or not Path(sys.argv[0]).is_file():
            script_path.write_text(sys.stdin.read())
        if not service.is_file():
            service.write_text(SERVICE.format(prefix=prefix))
        if not config.is_file():
            config.write_text(
                CONFIG.format(
                    path=os.getenv(
                        "FREVA_CONFIG_PATH", "/path/to/freva-config.toml"
                    )
                )
            )
        BootstrapConda(args.prefix, extra_pkgs)
        print(
            "Boot strapping success.\n"
            f"Check systemd {service} and {config} config file."
        )
        return

    if args.version:
        print(f"freva_setup.py version {__version__}")
        return

    if args.verify_checksum:
        actual = calculate_checksum(script_path)
        expected = SCRIPT_CHECKSUM.lower()
        print(f"[INFO] SHA256: {actual}")
        if actual == expected:
            print("[INFO] Checksum verified.")
            return
        else:
            raise SystemExit("[ERROR] Checksum mismatch!")

    if args.self_update:
        self_update(SCRIPT_URL, script_path)
        sys.exit(0)

    # Main execution
    log_dir = prefix / "log"
    log_dir.mkdir(exist_ok=True, parents=True)
    env = os.environ.copy()
    try:
        os.environ["SCRIPT_PATH"] = str(args.script_directory or "")
        with tempfile.NamedTemporaryFile(suffix=".conf") as temp_f:
            os.environ["FREVA_AUTOMATION_ENV_FILE"] = temp_f.name
            Path(temp_f.name).write_text(
                "\n".join(f"{e}={v}" for e, v in os.environ.items())
            )
            ps = PrefectServer(
                log_dir, port, web_port=args.web_port, endpoint=args.endpoint
            )
            ps.prep_server(args.script_directory.expanduser(), args.user)
            ps.start_prefect()
            register_prefect_deployment()
            if args.no_reverse_proxy is False:
                ps.start_caddy(args.cert_file, args.key_file)
            ps.linger()
    finally:
        os.environ = env
        try:
            ps.stop_server()
        except UnboundLocalError:
            pass


if __name__ == "__main__":
    main()



"""Script that prepares a new release of a version."""

import abc
import argparse
import json
import logging
import os
import re
import tempfile
from datetime import date
from functools import cached_property
from itertools import product
from pathlib import Path
from typing import Dict

import git
import requests
import tomli
from packaging.version import InvalidVersion, Version

# Set up logging
logging.basicConfig(
    format="%(name)s - %(levelname)s - %(message)s",
    datefmt="[%X]",
    level=logging.INFO,
)


logger = logging.getLogger("create-release")


class Release:
    """Abstract class for defining release jobs."""

    version_pattern: str = (
        r"__version__\s*=\s*['\"](\d+(\.\d+)*(-[a-zA-Z0-9]+)?)[\"']"
    )

    @abc.abstractmethod
    def __init__(
        self,
        package_name: str,
        repo_dir: str,
        branch: str = "main",
        search_path: str = ".",
        **kwargs: str,
    ) -> None:
        """Abstract init method."""

    @abc.abstractmethod
    def main(self) -> None:
        """Abstract main method."""


def cli(temp_dir: str) -> "Release":
    """Command line interface."""

    parser = argparse.ArgumentParser(
        description="Prepare the release of a package."
    )
    subparser = parser.add_subparsers(help="Available commands:")
    tag_parser = subparser.add_parser("tag", help="Create a new tag")
    deploy_parser = subparser.add_parser(
        "deploy", help="Update the version in the deployment repository"
    )
    rc_parser = subparser.add_parser(
        "rc", help="Check the release candidate version"
    )
    rc_parser.add_argument(
        "release-candidate", type=int, help="The release candidate number"
    )
    for _parser in (tag_parser, deploy_parser, rc_parser):
        _parser.add_argument("name", help="The name of the software/package.")
        _parser.add_argument(
            "-p", "--path", help="Set the search path.", type=str, default="."
        )
        _parser.add_argument(
            "-v",
            "--verbose",
            help="Enable debug mode.",
            action="store_true",
        )
    for _parser in (tag_parser, deploy_parser):
        _parser.add_argument(
            "-b",
            "--branch",
            help="Set the working branch",
            type=str,
            default="main",
        )
    deploy_parser.add_argument(
        "-s",
        "--services",
        help="Additional services and their versions.",
        type=str,
        nargs=2,
        action="append",
    )
    rc_parser.set_defaults(apply_func=ReleaseCandidate)
    tag_parser.set_defaults(apply_func=Tag)
    deploy_parser.set_defaults(apply_func=Bump)
    args = parser.parse_args()
    kwargs = {}
    if hasattr(args, "services"):
        kwargs = {s: v for (s, v) in args.services or ()}
    if args.verbose:
        logger.setLevel(logging.DEBUG)
    rc = getattr(args, "release-candidate", None)
    if rc is not None:
        return args.apply_func(rc, args.name, args.path)
    return args.apply_func(args.name, temp_dir, args.branch, args.path, **kwargs)


class Exit(Exception):
    """Custom error class for representing specific errors.

    Attributes:
        message (str): A human-readable message describing the error.
        code (int): An error code associated with the error.
    """

    def __init__(self, message: str, code: int = 1) -> None:
        """Initialize the CustomError instance with the given message and error code.

        Args:
            message (str): A human-readable message describing the error.
            code (int): An error code associated with the error.
        """
        super().__init__(message)
        self.message = message
        logger.exception(message)
        raise SystemExit(code)


class Bump(Release):
    """Bump the version."""

    owner = "freva-org"
    repo = "freva-admin"
    url = f"https://api.github.com/repos/{owner}/{repo}/pulls"

    def __init__(
        self,
        package_name: str,
        repo_dir: str,
        branch: str = "main",
        search_path: str = ".",
        **kwargs: str,
    ) -> None:
        self.extra_packages = {}
        self.version = os.environ.get("REPO_VERSION", "").strip("v")
        if not self.version:
            raise Exit("REPO_VERSION environment variable is not set", code=1)
        token = os.environ.get("DEPLOY_BOT_TOKEN", "")
        if not token:
            raise Exit("DEPLOY_BOT_TOKEN environment variable is not set", code=1)

        self.branch = branch
        self.token = token
        self.package_name = package_name
        self.repo_dir = Path(repo_dir)
        self.search_path = search_path
        self.repo_url = (
            f"https://x-access-token:{self.token}@github.com/"
            f"{self.owner}/{self.repo}.git"
        )
        logger.debug(
            "Cloning repository from %s with branch %s to %s",
            self.repo_url,
            self.repo_dir,
            branch,
        )
        self.git_repo = git.Repo.clone_from(
            self.repo_url, self.repo_dir, branch=branch
        )
        # Ensure commits are authored by the bot identity
        with self.git_repo.config_writer() as cw:
            cw.set_value(
                "user", "name", os.getenv("BOT_COMMIT_NAME", "freva-bot[bot]")
            )
            cw.set_value(
                "user",
                "email",
                os.getenv(
                    "BOT_COMMIT_EMAIL",
                    "freva-bot@users.noreply.github.com",
                ),
            )

    @property
    def repo_name(self) -> str:
        repo = git.Repo(search_parent_directories=True)
        name = repo.remotes.origin.url.split("/")[-1].replace(".git", "")
        return self.lookup.get(name, name)

    @property
    def lookup(self) -> Dict[str, str]:
        return {
            "freva-web": "web",
            "databrowserAPI": "databrowser",
            "freva": "core",
            "freva-nextgen": "freva_rest",
        }

    def update_whatsnew(self) -> None:
        """Update the what's new section for the current deployment version.

        Behaviour:
        - If a section for the current deploy_version already exists:
          * If a bullet for this service exists, replace its version.
          * Otherwise, append a new bullet for this service.
        - If no section exists yet, create one right after :titlesonly:.
        - Only change the minimal section of the file to reduce merge conflicts.
        """
        file = self.repo_dir / "docs" / "whatsnew.rst"
        text = file.read_text()

        # How this service is referred to in the text
        service = {v: k for k, v in self.lookup.items()}.get(
            self.package_name, self.package_name
        )

        version_str = self.deploy_version.public  # e.g. "2405.0.0"
        header_line = f"v{version_str}"
        underline = "~" * len(version_str)
        bullet_line = f"* Bumped version of {service} to {self.version}\n"

        # Regex to find the section header for this deploy version:
        # v2405.0.0
        # ~~~~~~~~
        header_pattern = rf"^v{re.escape(version_str)}\n(~+)\n"
        header_match = re.search(header_pattern, text, flags=re.MULTILINE)

        if header_match:
            # Section exists: work only inside this section
            section_start = (
                header_match.end()
            )  # index right after underline + newline

            # Find start of the next version header (or EOF)
            next_header_match = re.search(
                r"^v\d[\d\.]*\n~+\n", text[section_start:], flags=re.MULTILINE
            )
            if next_header_match:
                section_end = section_start + next_header_match.start()
            else:
                section_end = len(text)

            section_body = text[section_start:section_end]

            # Regex for "this service" bullet
            # e.g. "* Bumped version of freva_rest to 1.2.3"
            bullet_re = re.compile(
                rf"^\* Bumped version of {re.escape(service)} to .*$",
                flags=re.MULTILINE,
            )

            if bullet_re.search(section_body):
                # Replace existing bullet's version
                new_section_body = bullet_re.sub(
                    bullet_line.rstrip(),  # no trailing newline here
                    section_body,
                    count=1,
                )
                # Ensure we have a terminating newline
                new_section_body = new_section_body.rstrip() + "\n\n"
            else:
                # Append new bullet at the end of the section
                new_section_body = (
                    section_body.rstrip() + "\n" + bullet_line + "\n"
                )

            new_text = (
                text[:section_start] + new_section_body + text[section_end:]
            )
            file.write_text(new_text)
            return

        # No section for this deploy version yet: create one after :titlesonly:
        marker = ":titlesonly:"
        if marker in text:
            idx = text.index(marker) + len(marker)
            insertion = f"\n\n{header_line}\n{underline}\n{bullet_line}\n"
            new_text = text[:idx] + insertion + text[idx:]
        else:
            # Fallback: prepend a full section at the top
            new_text = (
                f"{marker}\n\n{header_line}\n{underline}\n{bullet_line}\n\n{text}"
            )

        file.write_text(new_text)

    @cached_property
    def deploy_version(self) -> str:
        """Get the deployment version."""
        version_tuple = self.old_deploy_version.release
        major = int(date.today().strftime("%y%m"))
        if major > version_tuple[0]:
            new_version = Version(f"{major}.0.0")
        else:
            new_version = Version(f"{version_tuple[0]}.{version_tuple[1]+1}.0")
        return new_version

    @cached_property
    def old_deploy_version(self) -> str:
        """Get the current deployment version."""
        file = Path(self.repo_dir / "src" / "freva_deployment" / "__init__.py")
        match = re.search(self.version_pattern, file.read_text())
        if not match:
            raise Exit("Could not find freva-deployment version", code=1)
        return Version(match.group(1))

    def _bump_service_version(self, service_file: Path) -> None:
        logger.debug("Bumping the service version.")
        versions = json.loads(service_file.read_text())
        old_version = versions[self.package_name]
        if old_version > self.version:
            raise Exit(
                "Service version is newer than bump version.: "
                "current version: {old_version}, new version: {self.version}",
                code=1,
            )
        versions[self.package_name] = self.version
        for service, vers in self.extra_packages.items():
            versions[service] = vers
        service_file.write_text(json.dumps(versions, indent=3))

    def _bump_deployment_version(self) -> None:
        file = Path(self.repo_dir / "src" / "freva_deployment" / "__init__.py")
        self._bump_service_version(file.parent / "versions.json")
        logger.debug("Looking for version")
        logger.debug("New version is %s", self.deploy_version.public)
        file_content = file.read_text().replace(
            self.old_deploy_version.public, self.deploy_version.public
        )
        file.write_text(file_content)

    def main(self) -> None:
        """Do the main work."""
        self.update_whatsnew()
        branch = f"bump-{self.package_name}-{self.version}"
        self._bump_deployment_version()
        logger.debug("Creating new branch %s", branch)
        self.git_repo.create_head(branch)
        self.git_repo.git.checkout(branch)
        self.git_repo.index.add("*")
        commit_message = f"Bump {self.package_name} version to {self.version}"
        self.git_repo.index.commit(commit_message)
        origin = self.git_repo.remote(name="origin")
        logger.debug("Submitting PR")
        origin.set_url(self.repo_url)
        origin.push(branch)
        self.submit_pr(branch)

    def submit_pr(self, branch: str) -> str:
        """Submit a PR on the deployment repo."""

        # Data for the pull request
        data = {
            "title": f"Bump {self.package_name} version to {self.version}",
            "head": branch,  # Source branch
            "base": self.branch,  # Target branch
            "body": (
                f"This PR auto bumps the version of {self.package_name}"
                f" to {self.version}. After the PR is merged you can create"
                " a new release of the deployment software by creating a"
                f" tag with the name v{self.version} or, better by following "
                "the release procedure:\n\n```console\ntox -e release\n```\n"
            ),
        }
        headers = {
            "Authorization": f"Bearer {self.token}",
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        }
        logger.info("Creating PR")
        response = requests.post(self.url, json=data, headers=headers)
        # Check if pull request was successfully created
        if response.status_code == 201:
            logger.debug("Pull request created successfully!")
        else:
            raise Exit("Failed to create pull request: {response.text}", code=0)


class Tag(Release):
    """Tag class."""

    def __init__(
        self,
        package_name: str,
        repo_dir: str,
        branch: str = "main",
        search_path: str = ".",
        version: str = "",
        **kwargs: str,
    ) -> None:
        self.branch = branch
        self.package_name = package_name
        self.repo_dir = Path(repo_dir)
        self.search_path = search_path
        logger.info(
            "Searching for packages/config with the name: %s", package_name
        )
        logger.debug("Reading current git config")
        self.git_config = (
            Path(git.Repo(search_parent_directories=True).git_dir) / "config"
        ).read_text()
        if os.environ.get("GIT_USER"):
            append = """[user]
    name = {user}
    email = {user}@users.noreply.github.com
    """.format(
                user=os.environ.get("GIT_USER")
            )
            self.git_config += "\n"
            self.git_config += append

    def tag_version(self) -> None:
        """Tag the latest git version."""

    @cached_property
    def repo_url(self) -> str:
        """Get the current git repo url."""
        repo = git.Repo(search_parent_directories=True)
        url = repo.remotes.origin.url
        token = os.environ.get("GITHUB_TOKEN")
        if token:
            url = url.replace("https://", f"https://{token}@")
        return url

    @cached_property
    def git_tag(self) -> Version:
        """Get the latest git tag."""
        logger.debug("Searching for the latest tag")
        repo = git.Repo(self.repo_dir)
        try:
            # Get the latest tag on the main branch
            return Version(
                repo.git.describe("--tags", "--abbrev=0", self.branch).lstrip("v")
            )
        except git.exc.GitCommandError:
            logger.debug("No tag found")
        except InvalidVersion:
            logger.debug("Tag found, but could not parse version")
        return Version("0.0.0")

    @property
    def version(self) -> Version:
        """Get the version of the current software."""
        logger.debug("Searching for software version.")
        pck_dirs = Path(self.search_path) / Path("src") / self.package_name, Path(
            self.search_path
        ) / Path("src") / self.package_name.replace("-", "_")
        files = [
            self.repo_dir / f[1] / f[0]
            for f in product(("_version.py", "__init__.py"), pck_dirs)
        ]
        files += [
            self.repo_dir / self.search_path / Path("package.json"),
            self.repo_dir / self.search_path / "pyproject.toml",
        ]
        for file in files:
            if file.is_file():
                if file.suffix == ".py":
                    match = re.search(self.version_pattern, file.read_text())
                    if match:
                        return Version(match.group(1))
                elif file.suffix == ".json":
                    content = json.loads(file.read_text())
                    if "version" in content:
                        return Version(content["version"])
                elif file.suffix == ".toml":
                    content = tomli.loads(file.read_text())
                    if "project" in content and "version" in content["project"]:
                        return Version(content["project"]["version"])
        raise ValueError("Could not find version")

    def _clone_repo_from_branch(self, branch: str = "main") -> None:
        logger.debug(
            "Cloning repository from %s with branch %s to %s",
            self.repo_url,
            self.repo_dir,
            branch,
        )
        git.Repo.clone_from(self.repo_url, self.repo_dir, branch=branch)
        (self.repo_dir / ".git" / "config").write_text(self.git_config)

    def _check_change_log_file(self) -> None:
        """Check if the current version was added to the change log file."""
        logger.debug("Checking for change log file.")
        version = self.version
        dev_rc_keywords = ["dev", "rc", "alpha", "beta"]
        if any(keyword in str(version) for keyword in dev_rc_keywords):
            return
        if not self._change_log_file.is_file():
            raise Exit(
                "Could not find change log file. "
                f"Create one first and push it to the {self.branch} branch."
            )
        if f"v{version}" not in self._change_log_file.read_text("utf-8"):
            raise Exit(
                "You need to add the version v{} to the {} change log file "
                "and push the update to the {} branch".format(
                    version,
                    self._change_log_file.relative_to(self.repo_dir),
                    self.branch,
                )
            )

    @cached_property
    def _change_log_file(self) -> Path:
        """Find the change log file."""
        for prefix, suffix in product(
            ("changelog", "whats-new", "whatsnew"), (".rst", ".md")
        ):
            for search_pattern in (prefix, prefix.upper()):
                glob_pattern = f"{search_pattern}{suffix}"
                logger.debug("Searching for %s", glob_pattern)
                for file in self.repo_dir.rglob(glob_pattern):
                    return file
        return Path(tempfile.mktemp())

    def main(self) -> None:
        """Tag a new git version."""
        self._clone_repo_from_branch(self.branch)
        cloned_repo = git.Repo(self.repo_dir)
        remote = cloned_repo.remote(name="origin")
        url = cloned_repo.remotes.origin.url
        token = os.environ.get("GITHUB_TOKEN")
        if token:
            url = url.replace("https://", f"https://{token}@")
            logger.info("Setting remote url to %s", url)
        remote.set_url(url)
        if self.version <= self.git_tag:
            raise Exit(
                "Tag version: {} is the same as current version {}"
                ", you need to bump the version number first and "
                "push the changes to the {} branch".format(
                    self.version,
                    self.git_tag,
                    self.branch,
                )
            )
        self._check_change_log_file()
        logger.info("Creating tag for version v%s", self.version)
        head = cloned_repo.head.reference
        message = f"Create a release for v{self.version}"
        try:
            cloned_repo.create_tag(f"v{self.version}", ref=head, message=message)
            cloned_repo.git.push("--tags")
        except git.GitCommandError as error:
            raise Exit("Could not create tag: {}".format(error))
        logger.info("Tags created.")


class ReleaseCandidate(Tag):
    def __init__(
        self,
        release_candidate: 0,
        repo_dir: str,
        search_path: str,
        **kwargs: str,
    ) -> None:
        self.release_candidate = release_candidate
        the_path = str(Path.cwd())
        _repo = git.Repo(the_path)
        self.branch = _repo.active_branch.name

        super().__init__(
            repo_dir, the_path, search_path=str(Path(search_path).absolute())
        )

    def main(self) -> None:
        """Get rc version."""
        the_version = Version(self.version.public + f"rc{self.release_candidate}")
        tag = self.get_latest_tag()
        if the_version <= tag:
            raise ValueError(
                "Tag version: {} is the smaller or equal as current version {}"
                ", you need to bump the version number first.".format(
                    tag,
                    the_version,
                )
            )
        print(the_version.public)

    def get_latest_tag(self) -> Version:
        """Get the latest git tag."""
        logger.debug("Searching for the latest tag")
        with tempfile.TemporaryDirectory() as td:
            git.Repo.clone_from(self.repo_url, td, branch="main")
            repo = git.Repo(td)
            try:
                # Get the latest tag on the main branch
                return Version(
                    repo.git.describe("--tags", "--abbrev=0", self.branch).lstrip(
                        "v"
                    )
                )
            except git.exc.GitCommandError:
                logger.debug("No tag found")
            except InvalidVersion:
                logger.debug("Tag found, but could not parse version")
            return Version("0.0.0")


if __name__ == "__main__":
    with tempfile.TemporaryDirectory() as temporary_dir:
        try:
            release = cli(temporary_dir)
            release.main()
        except Exception as error:
            if logger.getEffectiveLevel() == logging.DEBUG:
                raise
            raise Exit("An error occurred: {}".format(error)) from error



#!/usr/bin/env python3
import argparse
import os
import shlex
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Dict, List, Tuple

SYSTEMD_TMPL = dict(
    Unit=dict(
        Description="Start/Stop freva services containers",
        After="network-online.target",
    ),
    Service=dict(
        TimeoutStartSec="35s",
        TimeoutStopSec="35s",
        ExecStartPre='/bin/sh -c "{delete_command}"',
        ExecStart='/bin/sh -c "{container_cmd} {container_args}"',
        ExecStop='/bin/sh -c "{delete_command}"',
        Restart="on-failure",
        RestartSec=5,
        StartLimitBurst=5,
    ),
    Install=dict(WantedBy="default.target"),
)


def parse_args() -> Tuple[str, str, List[str], bool]:
    """Parse the commandline arguments."""

    app = argparse.ArgumentParser(
        prog=sys.argv[0], description="Create a new systemd unit."
    )
    app.add_argument("name", type=str, help="Set the container name")
    app.add_argument(
        "--enable",
        action="store_true",
        default=False,
        help="Enable this unit.",
    )
    app.add_argument("--requires", type=str, nargs="+", default=[])
    app.add_argument(
        "--gracious",
        action="store_true",
        default=False,
        help="Do not restart but only start the unit",
    )
    app.add_argument(
        "--print-unit-only",
        action="store_true",
        default=False,
        help="Only print the unit that would be created and started.",
    )
    args, other = app.parse_known_args()
    enable = args.enable
    if os.environ.get("DEBUG", "false").lower() == "true":
        enable = False
    return (
        " ".join(other),
        args.name,
        args.requires,
        enable,
        args.gracious,
        args.print_unit_only,
    )


def _parse_dict(tmp_dict: Dict[str, Dict[str, str]]) -> str:
    systemd_unit = ""
    for section, keys in tmp_dict.items():
        systemd_unit += "[{}]\n".format(section)
        for key, values in keys.items():
            systemd_unit += "{}={}\n".format(key, values)
    return systemd_unit


def load_unit(
    unit: str,
    content: str,
    enable: bool = True,
    gracious: bool = False,
    print_unit_only: bool = False,
) -> None:
    """Load a given systemd unit."""
    files = (
        "/etc/systemd/system/{}.service".format(unit),
        "~/.local/share/systemd/user/{}.service".format(unit),
    )
    if print_unit_only:
        print(content)
        return
    flags = ("", "--user")
    for file, flag in zip(files, flags):
        out_file = Path(file).expanduser()
        try:
            out_file.parent.mkdir(exist_ok=True, parents=True)
            with out_file.open(mode="w", encoding="utf-8") as f_obj:
                f_obj.write(content)
        except PermissionError:
            continue
        cmd = ["systemctl"]
        if flag:
            cmd += [flag]

        subprocess.run(cmd + ["daemon-reload"], check=True)
        if enable:
            subprocess.run(cmd + ["enable", unit], check=True)
        if gracious:
            subprocess.run(cmd + ["start", unit], check=True)
        else:
            subprocess.run(cmd + ["restart", unit], check=True)
        return


def get_container_cmd(args: str) -> Tuple[str, str]:
    """Get the correct container command for the system."""

    def _get_container_cmd(path: str) -> str:
        prefer = os.getenv("PREFER") or "podman"
        for cmd in (prefer,) + ("podman", "docker"):
            container_cmd = shutil.which(cmd, path=path)
            if container_cmd:
                return cmd
        raise ValueError("Docker or Podman must be installed")

    home_bin = os.path.join(os.path.expanduser("~"), ".local", "bin")
    path = (
        os.environ.get("PATH", "")
        + os.pathsep
        + "/usr/local/bin"
        + os.pathsep
        + home_bin
    )
    container_cmd = _get_container_cmd(path)
    args_list = shlex.split(args)
    if "compose" in args_list:
        proc = subprocess.run(
            [container_cmd, "compose"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False,
        )
        if "podman" in container_cmd:
            container_cmd = "podman-compose"
            args_list = [a for a in args_list if a != "compose"]
        if proc.returncode != 0:
            _ = args_list.pop(args_list.index("compose"))
            compose_cmd = f"{container_cmd}-compose"
            command = shutil.which(compose_cmd)
            if not command:
                for cmd in ("ensurepip", f"pip install {compose_cmd}"):
                    try:
                        subprocess.check_call(
                            shlex.split("python3 -m " + cmd),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                        )
                    except (subprocess.CalledProcessError, FileNotFoundError):
                        txt = (
                            f"{container_cmd} is available but not {compose_cmd}"
                            " which should be installed on the system."
                        )
                        print(txt, file=sys.stderr)
                        raise ValueError(txt)
            container_cmd = compose_cmd
    container_path = shutil.which(container_cmd) or container_cmd
    return container_path, " ".join(args_list)


def create_unit(
    args: str,
    unit: str,
    requires: List[str],
    enable: bool,
    gracious: bool,
    print_unit_only: bool = False,
) -> None:
    """Create the systemd unit."""
    container_cmd, container_args = get_container_cmd(args)
    cmd = shlex.split(args)
    if "compose" in cmd and "up" in cmd:
        new_cmd = []
        for word in cmd:
            if word == "up":
                new_cmd.append("down")
            elif word not in ("-d", "--detach"):
                new_cmd.append(word)
        _, delete_command = get_container_cmd(" ".join(new_cmd))
    else:
        _, delete_command = get_container_cmd("rm -f {}".format(unit))
    if delete_command:
        delete_command = "{} {}".format(container_cmd, delete_command)
    else:
        delete_command = ""
    if "docker" in container_cmd:
        SYSTEMD_TMPL["Unit"]["Requires"] = "docker.service"
        SYSTEMD_TMPL["Unit"]["After"] += " docker.service"
    # else:
    #    SYSTEMD_TMPL["Service"]["Environment"] = "PODMAN_USERNS=keep-id"
    for key in ("ExecStart",):
        SYSTEMD_TMPL["Service"][key] = SYSTEMD_TMPL["Service"][key].format(
            container_cmd=container_cmd,
            container_args=container_args,
            unit=unit,
        )
    for key in ("ExecStartPre", "ExecStop"):
        SYSTEMD_TMPL["Service"][key] = SYSTEMD_TMPL["Service"][key].format(
            delete_command=delete_command
        )
    for service in requires:
        for key in ("After", "Requires"):
            try:
                SYSTEMD_TMPL["Unit"][key] += " {}.service".format(service)
            except KeyError:
                SYSTEMD_TMPL["Unit"][key] = " {}.service".format(service)
    load_unit(
        unit,
        _parse_dict(SYSTEMD_TMPL),
        enable,
        gracious=gracious,
        print_unit_only=print_unit_only,
    )


if __name__ == "__main__":
    create_unit(*parse_args())
